# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.model.constants import EMPTY_ID, empty

__all__ = ["EMPTY_ID", "empty"]
